#ifndef _WHIRLYBIRD_H
#define _WHIRLYBIRD_H

#define KGSL_IOC_TYPE 0x09

struct kgsl_gpumem_get_info {
	unsigned long gpuaddr;
	unsigned int id;
	unsigned int flags;
	size_t size;
	size_t mmapsize;
	unsigned long useraddr;
	unsigned long __pad[4];
};

#define IOCTL_KGSL_GPUMEM_GET_INFO\
	_IOWR(KGSL_IOC_TYPE, 0x36, struct kgsl_gpumem_get_info)

struct kgsl_gpumem_alloc {
	unsigned long gpuaddr; /* output param */
	size_t size;
	unsigned int flags;
};

#define IOCTL_KGSL_GPUMEM_ALLOC \
	_IOWR(KGSL_IOC_TYPE, 0x2f, struct kgsl_gpumem_alloc)

#define CP_MEM_TO_MEM	   0x73

#define CP_TYPE7_PKT	(7 << 28)

static inline uint pm4_calc_odd_parity_bit(uint val) {
	return (0x9669 >> (0xf & ((val) ^
				  ((val) >> 4) ^ ((val) >> 8) ^ ((val) >> 12) ^
				  ((val) >> 16) ^ ((val) >> 20) ^ ((val) >> 24) ^
				  ((val) >> 28)))) & 1;
}

static inline uint cp_type7_packet(uint opcode, uint cnt) {
	return CP_TYPE7_PKT | ((cnt) << 0) |
	   (pm4_calc_odd_parity_bit(cnt) << 15) |
	   (((opcode) & 0x7F) << 16) |
	   ((pm4_calc_odd_parity_bit(opcode) << 23));
}

#define upper_32_bits(n) ((uint32_t)(((n) >> 16) >> 16))
#define lower_32_bits(n) ((uint32_t)(n))

static inline uint cp_gpuaddr(uint *cmds, uint64_t gpuaddr)
{
	uint *start = cmds;

	*cmds++ = lower_32_bits(gpuaddr);
	*cmds++ = upper_32_bits(gpuaddr);

	return cmds - start;
}

struct kgsl_command_object {
	uint64_t offset;
	uint64_t gpuaddr;
	uint64_t size;
	unsigned int flags;
	unsigned int id;
};

struct kgsl_gpu_command {
	uint64_t flags;
	uint64_t cmdlist;
	unsigned int cmdsize;
	unsigned int numcmds;
	uint64_t objlist;
	unsigned int objsize;
	unsigned int numobjs;
	uint64_t synclist;
	unsigned int syncsize;
	unsigned int numsyncs;
	unsigned int context_id;
	unsigned int timestamp;
};

#define IOCTL_KGSL_GPU_COMMAND \
	_IOWR(KGSL_IOC_TYPE, 0x4A, struct kgsl_gpu_command)

#define KGSL_CMDLIST_IB		 0x00000001U

static inline int kgsl_gpu_command(int fd, uint32_t ctx_id, uint64_t gpuaddr, uint32_t cmdsize) {
	struct kgsl_command_object *cmds;

	struct kgsl_gpu_command req = {
		.cmdsize = sizeof(struct kgsl_command_object),
		.numcmds = 1,
		.context_id = ctx_id,
	};
	size_t cmds_size;
	uint32_t i;

	cmds_size = sizeof(struct kgsl_command_object);

	cmds = (struct kgsl_command_object *) malloc(cmds_size);

	if (cmds == NULL) {
	return -1;
	}

	memset(cmds, 0, cmds_size);

	cmds->flags = KGSL_CMDLIST_IB;
	cmds->gpuaddr = gpuaddr;
	cmds->size = cmdsize;

	req.cmdlist = (unsigned long) cmds;

	return ioctl(fd, IOCTL_KGSL_GPU_COMMAND, &req);
}

struct kgsl_drawctxt_create {
	unsigned int flags;
	unsigned int drawctxt_id; /*output param */
};

#define IOCTL_KGSL_DRAWCTXT_CREATE \
	_IOWR(KGSL_IOC_TYPE, 0x13, struct kgsl_drawctxt_create)

static inline int kgsl_ctx_create0(int fd, uint32_t *ctx_id) {
	int ret; 

	struct kgsl_drawctxt_create req = {
		.flags = 0x00001812, /* low priority, rb 0 */
	};

	ret = ioctl(fd, IOCTL_KGSL_DRAWCTXT_CREATE, &req);

	if (ret)
		return ret;

	*ctx_id = req.drawctxt_id;

	return 0;
}

#endif
